(* Copyright 2003 Alan Post
 *
 * This file is part of mineml.
 *
 * mineml is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mineml is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mineml; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

type cell_state_t = Marked | Hidden | Open
exception Cant_hide_open
exception Cant_mark_open
exception Cant_open_marked
exception Dead of int * int
exception Won

type cell_t = {
    mutable state           : cell_state_t;
    mutable is_mine         : bool;
    mutable n_around        : int;
    mutable n_marked_around : int;
}

type t = {
    mutable cells      : cell_t array array;
            n_mines    : int;
            n_nonmines : int;
    mutable n_open     : int;
    mutable n_marked   : int;
    mutable dirties    : (int * int, unit) Hashtbl.t;
    mutable autoclear  : bool;
}

let width b = Array.length b.cells
let height b = Array.length b.cells.(0)
let n_surround b (x,y) = b.cells.(x).(y).n_around
let is_mine b (x,y) = b.cells.(x).(y).is_mine
let set_autoclear b ac = b.autoclear <- ac
let total_marked b = b.n_marked
let total_mines b = b.n_mines

let safe_do b (x,y) f = 
    if x >= 0 && x < (width b) && y >= 0 && y < (height b) then
        f b.cells.(x).(y) (x,y)

let do_around b (x,y) f = begin
    safe_do b (x-1,y-1) f;
    safe_do b (x-1,y  ) f;
    safe_do b (x-1,y+1) f;
    safe_do b (x  ,y-1) f;
    safe_do b (x  ,y+1) f;
    safe_do b (x+1,y-1) f;
    safe_do b (x+1,y  ) f;
    safe_do b (x+1,y+1) f;
end

let make ~width ~height ~n_mines ~autoclear =
    if n_mines > width * height then
        raise (Invalid_argument "too many mines")
    else
    let b = { cells = Array.init width
                        (fun _ -> Array.init height
                          (fun _ -> { state = Hidden;
                                      is_mine = false;
                                      n_around = 0;
                                      n_marked_around = 0 }));
              n_mines = n_mines;
              n_nonmines = width * height - n_mines;
              n_open = 0;
              n_marked = 0;
              dirties = Hashtbl.create 6;
              autoclear = autoclear } in
    let mines_left = ref n_mines in
    while !mines_left > 0 do
        let x = Random.int width
        and y = Random.int height in
        let cell = b.cells.(x).(y) in
        if not cell.is_mine then begin
             cell.is_mine <- true;
             do_around b (x,y) (fun c _ -> c.n_around <- c.n_around + 1 );
             decr mines_left
        end
    done;
    b

let state b (x,y) = b.cells.(x).(y).state

let rec clear_around b (x,y) =
    let c = b.cells.(x).(y) in
    if  c.state = Open && c.n_marked_around = c.n_around
      then do_around b (x,y) (fun c (x,y) -> if c.state = Hidden
                                             then set_state b (x,y) Open)
and set_state b (x,y) s = 
    let c = b.cells.(x).(y) in
    let old_s = c.state in
    if old_s <> s then
    match old_s, s with
      Open, Hidden -> raise Cant_hide_open
    | Open, Marked -> raise Cant_mark_open
    | Marked, Open -> raise Cant_open_marked
    | Hidden, Open when b.cells.(x).(y).is_mine -> raise (Dead (x,y))
    | _,_ -> ();
    c.state <- s;
    Hashtbl.replace b.dirties (x,y) ();
    begin
    let autoclear () = if b.autoclear then
                            do_around b (x,y)
                               (fun c (x,y) -> clear_around b (x,y)) in
    match old_s, s with
    | Marked, Hidden -> do_around b (x,y) (fun c _ -> c.n_marked_around <-
                                                      c.n_marked_around - 1);
                        b.n_marked <- b.n_marked - 1;
                        autoclear ()
    | Hidden, Marked -> b.n_marked <- b.n_marked + 1;
                        do_around b (x,y)
                               (fun c (x,y) ->
                                 c.n_marked_around <- c.n_marked_around + 1);
                        autoclear ()
    | Hidden, Open -> b.n_open <- b.n_open + 1;
                      if b.autoclear || c.n_around = 0 then clear_around b (x,y)
    | _,_ -> ()
    end;
    if ( b.n_open = b.n_nonmines ) then raise Won

let clear_dirties b = Hashtbl.clear b.dirties
let dirty_list b = Hashtbl.fold (fun k _ l -> k::l) b.dirties []

let as_string b =
    let buf = Buffer.create 100 in
    let go f = for y = 0 to (height b) - 1 do
                   for x = 0 to (width b) - 1 do
                       Buffer.add_string buf (f (x,y))
                   done;
                   Buffer.add_string buf "\n"
               done in
    go (fun (x,y) -> match state b (x,y) with
                       Open -> (match b.cells.(x).(y).n_around with
                                       0 -> " "
                                | _ as n -> string_of_int n)
                     | Marked -> "#"
                     | Hidden -> "-");
    Buffer.add_string buf "\n";
    go (fun (x,y) -> if b.cells.(x).(y).is_mine then "*"
                     else string_of_int b.cells.(x).(y).n_around);
    Buffer.contents buf

