(* Copyright 2003 Alan Post
 *
 * This file is part of mineml.
 *
 * mineml is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mineml is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mineml; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

open Board
open GdkKeysyms
open FacePixmaps
open CellPixmaps
open About

let img_state_from_state = function
      Hidden -> ImgHidden
    | Open   -> ImgOpen
    | Marked -> ImgMarked

let default_columns = 15
let min_columns = 10
let max_columns = 300
let default_rows = 10
let min_rows = 10
let max_rows = 200
let default_mine_density = 0.25
let cell_width = 20
let cell_height = 20

let border = 10
let button_border = 4

let next_columns = ref default_columns
let next_rows = ref default_rows
let next_mine_density = ref default_mine_density

let game_state = ref GamePlaying
let autoclear = ref true
let window = GWindow.window ~title:"mineml" ()
let _ = CellPixmaps.init window;
        FacePixmaps.init window;
        DigitPixmaps.init window

let top_vbox = GPack.vbox ~packing:window#add ()
let menubar = GMenu.menu_bar ~packing:top_vbox#pack ()
let vbox = let frame = GBin.frame ~shadow_type: `OUT
                                  ~packing: top_vbox#add () in
           GPack.vbox ~spacing: border
                      ~border_width: border
                      ~packing:frame#add ()

let statusbar = let frame = GBin.frame ~shadow_type: `IN
                            ~packing: vbox#pack () in
    GPack.hbox ~spacing: button_border
               ~border_width: button_border
               ~packing:frame#add ()

let frameme packing =
    let vbox = GPack.vbox ~packing () in
    let frame = GBin.frame ~shadow_type: `IN
                           ~packing: (vbox#pack ~expand:true ~fill:false) () in
    frame#add

let flags_left_disp = 
    Numbers.make
        ~packing:(frameme (statusbar#pack ~padding:(border - button_border)))
        ~digits:4

let face_button =
    GButton.button ~relief:`NONE
                   ~packing:(statusbar#pack ~expand:true
                                            ~fill:false) ()

let face_image = GMisc.image ~packing:face_button#add
                             ~pixmap:(FacePixmaps.get GamePlaying)#pixmap ()

let update_face () =
    face_image#set_pixmap (FacePixmaps.get !game_state)

let timer = Timer.make ~digits:4
    ~packing:(frameme (statusbar#pack ~padding:(border - button_border)))

let b = ref (make ~width:1 ~height:1 ~n_mines:0 ~autoclear:!autoclear)

let update_flags_left () =
    Numbers.set flags_left_disp ((total_mines !b) - (total_marked !b))

let eb = let frame = GBin.frame ~shadow_type: `IN
                                ~packing: vbox#add () in
         GBin.event_box ~packing: frame#add ()

let table = ref (GPack.table ~columns:1 ~rows:1 ~homogeneous:true
                             ~packing: eb#add ())

let cells = ref [|[||]|]

let fold_cells f init = snd (
    Array.fold_left (fun (x,acc) col ->
        (x+1, snd (Array.fold_left (fun (y,acc) img ->
                              (y+1, f acc img x y)) (0,acc) col)))
        (0,init) !cells)

let set_img_state x y surround st =
    (!cells).(x).(y)#set_pixmap (CellPixmaps.get st surround)

let redisp_dirties () =
    List.iter (fun (x,y) -> set_img_state x y (n_surround !b (x,y))
                                (img_state_from_state (state !b (x,y))))
              (dirty_list !b);
    (clear_dirties !b)

let end_game st dead =
    Timer.stop timer;
    redisp_dirties ();
    game_state := st;
    update_face ();
    fold_cells (fun _ _ x y ->
        match (state !b (x,y)), (is_mine !b (x,y))
        with
          Marked, false -> set_img_state x y 0 ImgWrong
        | Hidden, true -> set_img_state x y 0 ImgMine
        | _,_ -> ()) ();
    match dead with None -> ()
    | Some (x,y) -> set_img_state x y 0 ImgBlam

let table_callback ev =
    if    !game_state = GamePlaying
       && GdkEvent.get_type ev = `BUTTON_PRESS then begin
        begin
            let x = (int_of_float (GdkEvent.Button.x ev))/cell_width in
            let y = (int_of_float (GdkEvent.Button.y ev))/cell_height in
            try ((match GdkEvent.Button.button ev with
                     1 -> set_state !b (x,y) Open
                   | 2 -> clear_around !b (x,y)
                   | 3 -> set_state !b (x,y) (if state !b (x,y) = Marked
                                              then Hidden else Marked)
                   | _ -> ());
                  redisp_dirties ()) with
            | Won -> end_game GameWon None
            | Dead (x,y) -> end_game GameLost (Some (x,y))
            | _ -> ();
        end;
        update_flags_left ()
    end;
    true

let reinit () =
    game_state := GamePlaying;
    update_face ();
    (* let columns = columns_edit#value_as_int in
     * let rows    =    rows_edit#value_as_int in
     *)
    if (   !next_columns <> (width !b)
        || !next_rows <> (height !b)) then begin

        cells := Array.init (!next_columns) (fun _ ->
                 Array.init (!next_rows) (fun _ ->
                 GMisc.image ~pixmap:(CellPixmaps.get ImgHidden 0)#pixmap ()));
        eb#remove (!table)#coerce;
        table := GPack.table ~columns:!next_columns ~rows:!next_rows
                             ~homogeneous:true ~packing:eb#add ();
        fold_cells (fun _ img x y -> (!table)#attach ~left:x ~top:y img#coerce) ();
    end
    else begin
        fold_cells (fun _ _ x y -> set_img_state x y 0 ImgHidden) ();
    end;

    b := make ~width:!next_columns ~height:!next_rows
              ~n_mines:(int_of_float (   (float_of_int (  !next_columns
                                                        * !next_rows))
                                      *. !next_mine_density))
              ~autoclear:!autoclear;

    update_flags_left ();
    Timer.stop timer;
    Timer.start timer;

    (*
     *  Start the player off with an opening 
     *)
    let (opens, n_opens) = fold_cells (fun (l,n) _ x y ->
            if (n_surround !b (x,y)) = 0 && not (is_mine !b (x,y))
            then ((x,y)::l,n+1)
            else (l,n))
        ([],0)
    in 
    if n_opens > 0 then begin
        (try set_state !b (List.nth opens (Random.int n_opens)) Open
         with Won -> end_game GameWon None);
        redisp_dirties ()
    end

let mk_edit digits def min max step packing =
    let adjustment = GData.adjustment
        ~value: def
        ~lower: (float_of_int min)
        ~upper: (float_of_int max)
        ~step_incr: step ()
    in
    GEdit.spin_button ~adjustment ~rate: 1.2 ~numeric:true
                      ~digits ~packing ()

let launch_board_dialog () =
    let dialog = GWindow.dialog 
        ~parent: window
        ~destroy_with_parent: true
        ~title: "Board Setup"
        ~modal: true
        ~type_hint: `DIALOG
        ()
    in ignore (GMisc.label ~text: ("Changes will take effect at " ^
                                   "the start of the next board")
                           ~line_wrap: true 
                           ~packing: dialog#vbox#pack ());
       let tbl = GPack.table ~columns:2 ~rows:3 ~col_spacings:10
                             ~packing:dialog#vbox#pack () in
       ignore (GMisc.label ~text: "Columns:"
                           ~packing: (tbl#attach ~left:0 ~top:0) ());
       ignore (GMisc.label ~text: "Rows:"
                           ~packing: (tbl#attach ~left:0 ~top:1) ());
       ignore (GMisc.label ~text: "Mine Density:"
                           ~packing: (tbl#attach ~left:0 ~top:2) ());
       let c = mk_edit 0 (float_of_int !next_columns) min_columns max_columns
               1.0 (tbl#attach ~left:1 ~top:0) in
       let r = mk_edit 0 (float_of_int !next_rows) min_rows max_rows
               1.0 (tbl#attach ~left:1 ~top:1) in
       let m = mk_edit 3 !next_mine_density 0 1
               0.05 (tbl#attach ~left:1 ~top:2) in
       let ok = GButton.button ~stock:`OK
                               ~packing:dialog#action_area#pack () in
       ignore (ok#connect#clicked ~callback:
           (fun () -> next_columns := c#value_as_int;
                      next_rows := r#value_as_int;
                      next_mine_density := m#value;
                      dialog#destroy ()));
       let cancel = GButton.button ~stock:`CANCEL
                                   ~packing:dialog#action_area#pack () in
       ignore (cancel#connect#clicked ~callback:dialog#destroy);
       dialog#show ()

let _ =
    let factory = new GMenu.factory ~accel_path:"<MINEML>/" menubar in
    let accel_group = factory#accel_group in
    let game_menu = factory#add_submenu "Game" in
    let prefs_menu = factory#add_submenu "Prefs" in
    let about_menu = factory#add_submenu "About" in

    ignore (face_button#connect#clicked ~callback:reinit);
    ignore (eb#event#connect#button_press ~callback:table_callback);
    ignore (window#connect#destroy ~callback:GMain.Main.quit);
    let factory = new GMenu.factory ~accel_path:"<MINEML Game>/////" game_menu ~accel_group 
    in
    ignore (factory#add_item "New Game" ~key:_N ~callback:reinit);
    ignore (factory#add_separator ());
    ignore (factory#add_item "Quit" ~key:_Q ~callback:window#destroy);
    let factory = new GMenu.factory ~accel_path:"<MINEML Prefs>////" prefs_menu ~accel_group
    in
    let check_item = factory#add_check_item "Autoclear" ~active:!autoclear
    in
    ignore( check_item#connect#toggled
                ~callback:(fun () -> autoclear := check_item#active;
                                     set_autoclear !b !autoclear));
    ignore (factory#add_separator ());
    ignore (factory#add_item "Board Setup..." ~callback:launch_board_dialog);
    let factory = new GMenu.factory ~accel_path:"<MINEML About>////" about_menu ~accel_group
    in
    ignore (factory#add_item "About mineml" ~callback:(launch_about window));
    ignore (factory#add_item "License"      ~callback:(launch_license window));

    window#add_accel_group accel_group;
    Random.self_init ();
    reinit ();
    window#show ();
    Printexc.print GMain.Main.main ()
